Section predicate_logic.
Variable D: Set.
Variable c d e :D.
Variable P Q T: D-> Prop.


Theorem example_12 : (exists x, ~ P x) -> ~forall x, P x.
Proof.
intros.
intro H1.
destruct H as [y H].
 (* this replaces the assumption 
       "exists x : D, ~ P x" 
    by a fresh variable y : D and a hypothesis H : ~ P y
    also try 
       "elim H" and then "intros" and compare
  *)
apply H.
apply H1.
  (* This unifies the "P x" of "H : forall x : D, P x" with the goal "P y" 
     and finds the instantiation of y for x in the hypothesis 
   *)
Qed.

Theorem pred_023 : ((exists x, P x) -> forall x, Q x) -> 
  forall y, P y  -> Q y.
Proof.
intros.
apply H.
exists y.
  (* This says that we take "y" for the existentially quantified variable 
     in the goal
         "exists x : D, P x"
   *)
assumption.
Qed.

Theorem pred_008 : ~(forall x, P x)  -> ~ forall x, P x /\ Q x.
intros.
intro H1.
apply H.
intros x.
destruct (H1 x).
assumption.
Qed.

Theorem coq_pred_015 : 
  ~(forall x : D, P x \/ (Q x -> T x)) -> ~forall x : D, T x.
Proof.
intros.
intro H1.
apply H.
intro x.
right.
intro H2.
apply H1.
 (*! proof *)
Qed.


Theorem pred_025 : ~(forall x, P x /\ Q x) /\ (forall x, P x) -> 
  ~forall x, Q x.
Proof.
intros.
intro H1.
elim H.
intros H2 H3.
apply H2.
intro x.
split.
apply H3.
apply H1.
(*! proof *)
Qed.

(* Exercise 31 *)

(* Note how a binary relation on D is declared in Coq *)
Variable R : D -> D -> Prop.

Theorem pred_031 : (forall x, P x /\ Q x) -> forall x, P x \/ exists y, R x y.
Proof.
intros.
destruct (H x) as [H1 H2].
left; assumption.
(*! proof *)
Qed.

(* Exercise 36 *)

Theorem pred_036 : (exists x, forall y, ~R x y) -> ~forall x, R x x.
Proof.
intros.
destruct H.
intro H1.
unfold not in H.
apply H with x.
apply H1.
(*! proof *)
Qed.

Theorem pred_013 : (exists x, P x \/ Q x) -> (forall x, ~Q x) -> exists x, P x.
intros.
Admitted.

Theorem pred_035 : (forall y, Q y -> ~exists x, P x) /\ (forall x, P x) -> forall y, ~Q y.
Proof.
Admitted.

(* more difficult *)
Theorem pred_016 : (forall x, P x \/ R x x) -> (forall x, P x -> exists y, R x y /\ R y x) -> forall x, exists y, R x y.
intros.
destruct (H x).
destruct (H0 x H1) as [y0 Hy0].
exists y0.
destruct Hy0;assumption.
exists x; assumption.
Qed.


Theorem pred_067 : (forall x, ~P x) -> ~exists x, P x.
Proof.
Admitted.

(* Predicate logic with equality *)

Theorem example_14 : forall x y, R y y /\ x = y -> R y x.
Proof.
intros.
destruct H as [H1 H2].
rewrite H2.
 (* This replaces x by y in the goal
    Also try
      "rewrite <- H2"
    and see what happens
    Also try
      "rewrite <- H2 in H1"
    and see what happens
  *)
assumption.
Qed.

Variable W : D -> D -> D -> Prop.

Theorem example_17 : forall x y, W x x y /\ x = y -> W x y y.
Proof.
intros.
destruct H.
rewrite H0.
rewrite H0 in H.
assumption.
Qed.

Theorem pred_059 : (forall x:D, forall y, x = y) -> (exists x, P x) -> forall x, P x.
intros.
destruct H0.
rewrite (H x x0).
assumption.
Qed.

Theorem pred_058 : (forall x:D, forall y, x = y) /\ P d -> P e.
Admitted.

Section pred_080.

Hypothesis Domain : forall x, (x = c \/ x = d \/ x = e).

Theorem pred_080 : (forall x, P x) -> P e /\ P d /\ P c.
Proof.
(*! pred_proof *)
Admitted.


End pred_080.

(*CHALLENGE PROBLEM
    This theorem shows that if we have a transitive
   relation on a 3-element set, and if each element has
   a successor, then some element must be reflexive for
   this relation *)

Section pred_078.

Hypothesis Domain : exists x1, exists x2, exists x3, forall x:D, (x = x1 \/ x = x2 \/ x = x3).
Hypothesis Successor : forall x, exists y, R x y.
Hypothesis Transitive : forall x, forall y, forall z, R x y /\ R y z -> R x z.

Theorem pred_078 : exists x, R x x.
Proof.
destruct Domain as [x1 [x2 [x3 H]]].
destruct (Successor x1) as [x4 Hx4].
destruct (H x4).
exists x1; rewrite H0 in Hx4 ; apply Hx4.
destruct H0.

rewrite H0 in Hx4.
destruct (Successor x2) as [x5 Hx5].
destruct (H x5).
rewrite H1 in Hx5.
exists x1.
apply Transitive with x2.
intuition.
destruct H1.
rewrite H1 in Hx5.
exists x2; intuition.
rewrite H1 in Hx5.
destruct (Successor x3).
destruct (H x).
rewrite H3 in H2.
exists x1.
apply Transitive with x3.
split.
apply Transitive with x2.
intuition.
intuition.
destruct H3.
rewrite H3 in H2.
exists x2.
apply Transitive with x3.
intuition.
rewrite H3 in H2.
exists x3.
intuition.

rewrite H0 in Hx4.
destruct (Successor x3) as [x5 Hx5].
destruct (H x5).
rewrite H1 in Hx5.
exists x1.
apply Transitive with x3.
intuition.
destruct H1.
rewrite H1 in Hx5.
destruct (Successor x2).
destruct (H x).
rewrite H3 in H2.
exists x1.
apply Transitive with x3.
split.
intuition.
apply Transitive with x2.
intuition.
destruct H3.
rewrite H3 in H2.
exists x2.
intuition.
rewrite H3 in H2.
exists x3.
apply Transitive with x2.
intuition.
rewrite H1 in Hx5.
exists x3.
intuition.
Qed.


End pred_078.


End predicate_logic.